from typing import Union

from pydantic import BaseModel, AnyHttpUrl, FilePath
from pydantic_settings import BaseSettings, SettingsConfigDict


class AwtrixHttpConfig(BaseModel):
    base_url: AnyHttpUrl
    username: str = None
    password: str = None
    verify_ssl: Union[bool, FilePath] = False


class AwtrixLightHttpClientSettings(BaseSettings):
    awtrix: AwtrixHttpConfig

    model_config = SettingsConfigDict(
        env_prefix="AWTRIX_HTTP_CLIENT_", env_file=".env", env_file_encoding="utf-8"
    )



from pydantic import BaseModel
from enum import Enum


class Palette(str, Enum):
    CLOUD = "Cloud"
    LAVA = "Lava"
    OCEAN = "Ocean"
    FOREST = "Forest"
    STRIPE = "Stripe"
    PARTY = "Party"
    HEAT = "Heat"
    RAINBOW = "Rainbow"


class EffectType(str, Enum):
    FADE = "Fade"
    MOVINGLINE = "MovingLine"
    BRICKBREAKER = "BrickBreaker"
    PINGPONG = "PingPong"
    RADAR = "Radar"
    CHECKERBOARD = "Checkerboard"
    FIREWORKS = "Fireworks"
    PLASMACLOUD = "PlasmaCloud"
    RIPPLE = "Ripple"
    SNAKE = "Snake"
    PACIFICA = "Pacifica"
    THEATERCHASE = "TheaterChase"
    PLASMA = "Plasma"
    MATRIX = "Matrix"
    SWIRLIN = "SwirlIn"
    SWIRLOUT = "SwirlOut"
    LOOKINGEYES = "LookingEyes"
    TWINKLINGSTARS = "TwinklingStars"
    COLORWAVES = "ColorWaves"


class EffectSetting(BaseModel):
    speed: int
    palette: Palette
    blend: bool



from typing import List, Literal

from pydantic_extra_types.color import Color


def convert_color_to_hex(color: Color) -> str:
    return as_hex(color, format="long").upper()


def convert_colors_to_hex(colors: List[Color]) -> str:
    return [convert_color_to_hex(color) for color in colors]


def as_hex(c: Color, format: Literal["short", "long"] = "short") -> str:
    """
    # https://github.com/pydantic/pydantic-extra-types/blob/main/pydantic_extra_types/color.py#L141
    Will be removed when pydantic-extra-type will release a new version
    """
    values = [round(c * 255) for c in c._rgba[:3]]
    if c._rgba.alpha is not None:
        values.append(round(c._rgba.alpha * 255))

    as_hex = "".join(f"{v:02x}" for v in values)
    if format == "short" and all(
        c in {int(c * 2, 16) for c in "0123456789abcdef"} for c in values
    ):
        as_hex = "".join(as_hex[c] for c in range(0, len(as_hex), 2))
    return "#" + as_hex



from pydantic import BaseModel, model_validator, field_serializer
from pydantic_extra_types.color import Color

from .utils import convert_color_to_hex


class Moodlight(BaseModel):
    brightness: int = None
    kelvin: int = None
    color: Color = None

    @model_validator(mode="after")
    def check_constraint_blink_text(self) -> "Moodlight":
        if self.kelvin and self.color:
            raise ValueError("kelvin and color can't be set together")
        return self

    @field_serializer("color")
    def convert_color_to_hex(v: Color) -> str:
        if v:
            return convert_color_to_hex(v)
        return v



from typing import List, Union, Any, Optional
from enum import IntEnum

from pydantic import BaseModel, Field, model_validator, field_serializer
from pydantic.networks import Annotated, Url, UrlConstraints
from pydantic_extra_types.color import Color

from .effect import EffectType, EffectSetting
from .utils import convert_color_to_hex, convert_colors_to_hex


class TextCase(IntEnum):
    GLOBAL = 0
    FORCE_UPPERCASE = 1
    SHOW_AS_IT_SEND = 2


class PushIcon(IntEnum):
    NOT_MOVING = 0
    MOVING_ONE_TIME = 1
    MOVING = 2


class Dp(BaseModel):
    x: int
    y: int
    cl: Color

    @field_serializer(
        "cl",
    )
    def convert_color_to_int(v: Color) -> str:
        if v:
            return convert_color_to_hex(v)
        else:
            return v


class Dl(BaseModel):
    x0: int
    y0: int
    x1: int
    y1: int
    cl: Color

    @field_serializer(
        "cl",
    )
    def convert_color_to_int(v: Color) -> str:
        if v:
            return convert_color_to_hex(v)
        else:
            return v


class Dr(BaseModel):
    x: int
    y: int
    w: int
    h: int
    cl: Color

    @field_serializer(
        "cl",
    )
    def convert_color_to_int(v: Color) -> str:
        if v:
            return convert_color_to_hex(v)
        else:
            return v


class Df(BaseModel):
    x: int
    y: int
    w: int
    h: int
    cl: Color

    @field_serializer(
        "cl",
    )
    def convert_color_to_int(v: Color) -> str:
        if v:
            return convert_color_to_hex(v)
        else:
            return v


class Dc(BaseModel):
    x: int
    y: int
    r: int
    cl: Color

    @field_serializer(
        "cl",
    )
    def convert_color_to_int(v: Color) -> str:
        if v:
            return convert_color_to_hex(v)
        else:
            return v


class Dfc(BaseModel):
    x: int
    y: int
    r: int
    cl: Color

    @field_serializer(
        "cl",
    )
    def convert_color_to_int(v: Color) -> str:
        if v:
            return convert_color_to_hex(v)
        else:
            return v


class Dt(BaseModel):
    x: int
    y: int
    t: int
    cl: Color

    @field_serializer(
        "cl",
    )
    def convert_color_to_int(v: Color) -> str:
        if v:
            return convert_color_to_hex(v)
        else:
            return v


class Db(BaseModel):
    x: int
    y: int
    r: int
    cl: Any

    # cl: Union[bytes, bytearray, memoryview] need to generate pydantic shema, leaving it for now
    @field_serializer(
        "cl",
    )
    def convert_color_to_int(v: Color) -> str:
        if v:
            return convert_color_to_hex(v)
        else:
            return v


class LifeTimeMode(IntEnum):
    DELETE = 0
    STALE = 1


class Fragment(BaseModel):
    t: str
    c: Color

    @field_serializer(
        "c",
    )
    def convert_color_to_int(v: Color) -> str:
        if v:
            return convert_color_to_hex(v)
        else:
            return v


class BaseApplication(BaseModel):
    text: Optional[Union[str, List[Fragment]]] = None
    textCase: Optional[TextCase] = None
    topText: Optional[bool] = None
    textOffset: Optional[int] = Field(default=None, ge=0)
    center: Optional[bool] = None
    color: Optional[Color] = None
    gradient: Optional[List[Color]] = Field(default=None, min_length=2, max_length=2)
    blinkText: Optional[int] = None
    fadeText: Optional[int] = None
    background: Optional[Color] = None
    rainbow: Optional[bool] = None
    icon: Optional[str] = None
    pushIcon: Optional[PushIcon] = None
    repeat: Optional[int] = Field(default=None, ge=-1)
    duration: Optional[int] = Field(default=None, ge=1)
    bar: Optional[List[int]] = Field(default=None, max_length=16)
    line: Optional[List[int]] = Field(default=None, max_length=16)
    autoscale: Optional[bool] = None
    progress: Optional[int] = Field(default=None, ge=-1, le=100)
    progressC: Optional[Color] = None
    progressBC: Optional[Color] = None
    draw: Optional[List[Union[Dp, Dl, Dr, Df, Dc, Dfc, Dt, Db]]] = None
    noScroll: Optional[bool] = None
    scrollSpeed: Optional[int] = Field(default=None, ge=0, le=100)
    effect: Optional[EffectType] = None
    effectSettings: Optional[EffectSetting] = None

    class ConfigDict:
        use_enum_values = True

    @model_validator(mode="after")
    def check_constraint_blink_text(self) -> "BaseApplication":
        if self.blinkText and (self.gradient or self.rainbow):
            raise ValueError(
                "blink text can be set only if gradient and rainbow are not used"
            )
        return self

    @model_validator(mode="after")
    def check_constraint_fade_text(self) -> "BaseApplication":
        if self.fadeText and (self.gradient or self.rainbow):
            raise ValueError(
                "fade text can be set only if gradient and rainbow are not used"
            )
        return self

    @model_validator(mode="after")
    def check_constraint_bar(self) -> "BaseApplication":
        if self.icon and self.bar and len(self.bar) > 11:
            raise ValueError("bar can have only 11 values with icon")
        return self

    @model_validator(mode="after")
    def check_constraint_line(self) -> "BaseApplication":
        if self.icon and self.line and len(self.line) > 11:
            raise ValueError("line can have only 11 values with icon")
        return self

    @field_serializer("color", "gradient", "background", "progressC", "progressBC")
    def convert_color_to_hex(v: Union[List[Color], Color]) -> Union[List, str]:
        if isinstance(v, List):
            return convert_colors_to_hex(v)
        else:
            return convert_color_to_hex(v)


class CustomApplication(BaseApplication):
    pos: Optional[int] = Field(default=None, ge=0)
    lifetime: Optional[int] = Field(default=None, ge=0)
    lifetimeMode: Optional[LifeTimeMode] = None
    save: Optional[bool] = None


CLIENT_TYPE = Annotated[
    Url, UrlConstraints(allowed_schemes=["http", "https", "mqtt", "mqtts"])
]


class Notification(BaseApplication):
    hold: Optional[bool] = None
    sound: Optional[str] = None
    rtttl: Optional[str] = None
    loopSound: Optional[bool] = None
    stack: Optional[bool] = None
    wakeup: Optional[bool] = None
    clients: Optional[List[CLIENT_TYPE]] = None

    @model_validator(mode="after")
    def check_constraint_sound_rtttl(self) -> "Notification":
        if self.sound and self.rtttl:
            raise ValueError("sound and rtttl can't be set together")
        return self

    @field_serializer("clients")
    def convert_url_to_str(clients: CLIENT_TYPE) -> List[str]:
        return [str(client) for client in clients]



from typing import Dict, Any, List, Union, Literal
from contextlib import asynccontextmanager
from typing import AsyncIterator

from httpx import AsyncClient
from pydantic_extra_types.color import Color

from .http_settings import AwtrixLightHttpClientSettings
from .models.stat import Stats
from .models.effect import EffectType
from .models.transition import TransitionType
from .models.loop import Loop
from .models.screen import Screen
from .models.moodlight import Moodlight
from .models.application import CustomApplication, Notification
from .models.setting import Settings

from .models.utils import as_hex


class AwtrixLightHttpClientError(BaseException):
    def __init__(self, status_code: int, content: str, *args: object) -> None:
        super().__init__(*args)
        self.status_code = status_code
        self.content = content


class AwtrixLightHttpClient:
    def __init__(self, client: AsyncClient) -> None:
        self._client = client

    async def _make_request(
        self,
        method: str,
        url: str,
        params: Dict[Any, Any] = None,
        data: Dict[Any, Any] = None,
        json_data: Dict[Any, Any] = None,
    ):
        r = await self._client.request(
            method, url, params=params, data=data, json=json_data
        )

        if not r.is_success:
            raise AwtrixLightHttpClientError(
                status_code=r.status_code, content=r.content
            )
        return r

    async def get_stats(self) -> Stats:
        """
        General device stats (e.g., battery, RAM)
        """
        response = (await self._make_request("GET", "stats")).json()

        return Stats(**response)

    async def get_effects(self) -> List[EffectType]:
        """
        List of all effects
        """
        response = (await self._make_request("GET", "effects")).json()

        return [EffectType(e) for e in response]

    async def get_transitions(self) -> List[TransitionType]:
        """
        List of all transition effects
        """
        response = (await self._make_request("GET", "transitions")).json()

        return [(TransitionType[t.upper()]) for t in response]

    async def get_loops(self) -> Loop:
        """
        List of all apps in the loop
        """
        response = (await self._make_request("GET", "loop")).json()

        sorted_apps = dict(sorted(response.items(), key=lambda item: item[1]))

        return Loop(loops=[app for app in sorted_apps.keys()])

    async def get_screen(self) -> Screen:
        """
        Retrieve the current matrix screen as an array of 24 bit colors
        """
        response = (await self._make_request("GET", "screen")).json()

        return Screen(matrix=response)

    async def set_power(self, power: bool) -> None:
        """
        Toggle the matrix on or off
        """
        await self._make_request("POST", "power", json_data={"power": power})

    async def set_sleep(self, seconds: int) -> None:
        """
        Send the board in deep sleep mode (turns off the matrix as well), good for saving battery life
        """
        await self._make_request("POST", "sleep", json_data={"sleep": seconds})

    async def set_sound(self, sound: str) -> None:
        """
        Play a RTTTL sound from the MELODIES folder
        """
        await self._make_request("POST", "sound", json_data={"sound": sound})

    async def set_rtttl(self, rtttl: str) -> None:
        """
        Play a RTTTL sound from a given RTTTL string
        """
        await self._make_request("POST", "rtttl", json_data={"rtttl": rtttl})

    async def set_moodlight(self, moodlight: Moodlight) -> None:
        """
        Set the entire matrix to a custom color or temperature
        To disable moodlight pass `Moodlight()`
        """
        await self._make_request(
            "POST", "moodlight", json_data=moodlight.model_dump(exclude_none=True)
        )

    async def set_indicator(
        self,
        indicator: Literal[1, 2, 3],
        color: Color,
        blink: int = None,
        fade: int = None,
    ) -> None:
        """
        Colored indicators serve as small notification signs displayed on specific areas of the screen:
            Upper right corner: Indicator 1
            Right side: Indicator 2
            Lower right corner: Indicator 3
        To hide the indicators pass black as Color
        Blink is in milliseconds
        Fade is in milliseconds
        """
        if blink and fade:
            raise ValueError("fade and blink can't be set together")

        json_data = {"color": as_hex(color, format="long").upper()}

        if blink:
            json_data["blink"] = blink

        if fade:
            json_data["fade"] = fade

        await self._make_request(
            "POST",
            f"indicator{indicator}",
            json_data=json_data,
        )

    async def set_custom_application(
        self,
        name: str,
        custom_application: Union[CustomApplication, List[CustomApplication], None],
    ) -> None:
        """
        Set custom app or a list of custom app
        When erasing apps, AWTRIX doesn't match the exact app name. Instead, it identifies apps that begin with the specified name.
        To expunge all associated apps, send application=None. For example for name=test. This action will remove test0, test1, and so on.
        To eradicate a single app, direct the command to, for instance, test1
        """
        if isinstance(custom_application, CustomApplication):
            json_data = custom_application.model_dump(exclude_none=True)
        else:
            json_data = [
                app.model_dump(exclude_none=True) for app in custom_application
            ]

        await self._make_request(
            "POST", "custom", params={"name": name}, json_data=json_data
        )

    async def notify(self, notification: Notification) -> None:
        """
        One-Time Notification
        """
        await self._make_request(
            "POST", "notify", json_data=notification.model_dump(exclude_none=True)
        )

    async def dismiss_notification(self) -> None:
        """
        Easily dismiss a notification that was configured with "hold": true
        """
        await self._make_request("POST", "notify/dismiss")

    async def next_app(self) -> None:
        """
        Navigate to the next app
        """
        await self._make_request("POST", "nextapp")

    async def previous_app(self) -> None:
        """
        Navigate to the previous app
        """
        await self._make_request("POST", "previousapp")

    async def switch_app(self, name: str) -> None:
        """
        Directly transition to a desired app using its name
        """
        await self._make_request("POST", "switch", json_data={"name": name})

    async def get_settings(self) -> Settings:
        """
        You can initiate the firmware update either through the update button in HA or using the following
        """
        return Settings(**(await self._make_request("GET", "settings")).json())

    async def set_settings(self, s: Settings) -> None:
        """
        You can initiate the firmware update either through the update button in HA or using the following
        """
        await self._make_request(
            "POST", "settings", json_data=s.model_dump(exclude_none=True)
        )

    async def update(self) -> None:
        """
        You can initiate the firmware update either through the update button in HA or using the following
        """
        await self._make_request("POST", "doupdate")

    async def reboot(self) -> None:
        """
        If you need to restart the Awtrix
        """
        await self._make_request("POST", "reboot")

    async def erase(self) -> None:
        """
        WARNING: This action will format the flash memory and EEPROM but will not modify the WiFi Settings. It essentially serves as a factory reset.
        """
        await self._make_request("POST", "erase")

    async def reset_settings(self) -> None:
        """
        WARNING: This action will reset all settings from the settings API. It does not reset the flash files and WiFi Settings.
        """
        await self._make_request("POST", "resetSettings")


@asynccontextmanager
async def get_awtrix_http_client() -> AsyncIterator[AwtrixLightHttpClient]:
    settings = AwtrixLightHttpClientSettings()

    auth = None
    if settings.awtrix.username and settings.awtrix.password:
        auth = (
            settings.awtrix.username,
            settings.awtrix.password,
        )

    async with AsyncClient(
        base_url=f"{settings.awtrix.base_url}api",
        auth=auth,
        verify=settings.awtrix.verify_ssl,
    ) as client:
        yield AwtrixLightHttpClient(client)



from typing import Any
from typing import Literal, Union

from pydantic_extra_types.color import Color
from pydantic import BaseModel, Field, field_validator, field_serializer

from .transition import TransitionType
from .utils import convert_color_to_hex, as_hex


class Settings(BaseModel):
    ATIME: int = Field(default=None, ge=0)
    TEFF: TransitionType = None
    TSPEED: int = Field(default=None, ge=0)
    TCOL: Union[Color, int] = None
    TMODE: int = Field(default=None, ge=0, le=4)
    CHCOL: Union[Color, int] = None
    CBCOL: Union[Color, int] = None
    CTCOL: Union[Color, int] = None
    WD: bool = None
    WDCA: Union[Color, int] = None
    WDCI: Union[Color, int] = None
    BRI: int = Field(default=0, ge=0, le=255)
    ABRI: bool = None
    ATRANS: bool = None
    CCORRECTION: Union[Color, str] = None
    CTEMP: Union[Color, str] = None
    TFORMAT: Literal[
        "%H:%M:%S",
        "%l:%M:%S",
        "%H:%M",
        "%H %M",
        "%l:%M",
        "%l %M",
        "%l:%M %p",
        "%l %M %p",
    ] = None
    DFORMAT: Literal[
        "%d.%m.%y",
        "%d.%m",
        "%y-%m-%d",
        "%m-%d",
        "%m/%d/%y",
        "%m/%d",
        "%d/%m/%y",
        "%d/%m",
        "%m-%d-%y",
    ] = None
    SOM: bool = None
    CEL: bool = None
    MAT: int = None
    SOUND: bool = None
    GAMMA: float = None
    BLOCKN: bool = None
    UPPERCASE: bool = None
    TIME_COL: Union[Color, Literal[0]] = None
    DATE_COL: Union[Color, Literal[0]] = None
    TEMP_COL: Union[Color, Literal[0]] = None
    HUM_COL: Union[Color, Literal[0]] = None
    BAT_COL: Union[Color, Literal[0]] = None
    SSPEED: int = Field(default=None, ge=0, le=100)
    TIM: bool = None
    DAT: bool = None
    HUM: bool = None
    TEMP: bool = None
    BAT: bool = None
    MATP: bool = None

    class ConfigDict:
        use_enum_values = True

    @field_validator("CCORRECTION", "CTEMP", mode="before")
    @classmethod
    def convert_str_to_color(cls, v: Union[Color, str]) -> Color:
        if isinstance(v, str):
            return Color(v)
        else:
            return v

    @field_validator("TCOL", "CHCOL", "CBCOL", "CTCOL", "WDCA", "WDCI", mode="before")
    @classmethod
    def convert_integer_to_color(cls, v: Union[Color, int]) -> Color:
        if isinstance(v, int):
            return Color("{0:06X}".format(v))
        else:
            return v

    @field_serializer("CCORRECTION", "CTEMP")
    def convert_color_to_hex(v: Union[Color, str]) -> str:
        if isinstance(v, Color):
            return convert_color_to_hex(v)
        else:
            return v

    @field_serializer(
        "TCOL",
        "CHCOL",
        "CBCOL",
        "CTCOL",
        "WDCA",
        "WDCI",
        "TIME_COL",
        "DATE_COL",
        "TEMP_COL",
        "HUM_COL",
        "BAT_COL",
    )
    def convert_color_to_int(v: Any) -> int:
        if isinstance(v, Color):
            return int(as_hex(v, format="long")[1:], 16)
        else:
            return v



from enum import IntEnum


class TransitionType(IntEnum):
    RANDOM = 0
    SLIDE = 1
    DIM = 2
    ZOOM = 3
    ROTATE = 4
    PIXELATE = 5
    CURTAIN = 6
    RIPPLE = 7
    BLINK = 8
    RELOAD = 9
    FADE = 10



from pydantic import BaseModel


class Stats(BaseModel):
    bat: int
    bat_raw: int
    type: int
    lux: int
    ldr_raw: int
    ram: int
    bri: int
    temp: int
    hum: int
    uptime: int
    wifi_signal: int
    messages: int
    version: str
    indicator1: bool
    indicator2: bool
    indicator3: bool
    app: str
    uid: str
    matrix: bool
